"""Test the Marshmallow schemas for the institution models."""
from collections import OrderedDict

import gator.core.models.institution as models
import gator.core.schemas.institution as schemas


class TestInstitutionSchema:
    """Test the :class:`gator.core.schemas.institution.InstitutionSchema` class.

    Class Attributes:
        FACULTY_INSTITUTION: A faculty institution to test with.
        CAMPUS_INSTITUTION: A campus institution to test with.
        UNI_INSTITUTION: A university institution to test with.
        DUMPED_FACULTY_INSTITUTION: The dumped data for `FACULTY_INSTITUTION`.
        DUMPED_CAMPUS_INSTITUTION: The dumped data for `CAMPUS_INSTITUTION`.
        DUMPED_UNI_INSTITUTION: The dumped data for `UNI_INSTITUTION`.
    """

    FACULTY_INSTITUTION: models.Institution = models.Institution(
        code='ARTSC',
        name='Faculty of Arts and Science',
        type='faculty',
        parent=None,
    )
    CAMPUS_INSTITUTION: models.Institution = models.Institution(
        code='UTSG',
        name='St. George Campus',
        type='campus',
    )
    UNI_INSTITUTION: models.Institution = models.Institution(
        code='UofT',
        name='University of Toronto',
        type='university',
    )
    DUMPED_FACULTY_INSTITUTION: dict = {
        'code': 'ARTSC',
        'name': 'Faculty of Arts and Science',
        'type': 'faculty',
        'parent': 'UTSG'
    }
    DUMPED_CAMPUS_INSTITUTION: dict = {
        'code': 'UTSG',
        'name': 'St. George Campus',
        'type': 'campus',
        'parent': 'UofT',
        'sub_institutions': ['ARTSC']
    }
    DUMPED_UNI_INSTITUTION: dict = {
        'code': 'UofT',
        'name': 'University of Toronto',
        'type': 'university',
        'sub_institutions': ['UTSG']
    }

    def setup_class(self):
        """Set up the tests."""
        self.schema = schemas.InstitutionSchema()
        # Setup the parent-child relationships
        self.FACULTY_INSTITUTION.parent = self.CAMPUS_INSTITUTION
        self.CAMPUS_INSTITUTION.parent = self.UNI_INSTITUTION
        self.UNI_INSTITUTION.parent = None

        # Setup the sub-institution relationships
        self.FACULTY_INSTITUTION.sub_institutions = []
        self.CAMPUS_INSTITUTION.sub_institutions.append(self.FACULTY_INSTITUTION)
        self.UNI_INSTITUTION.sub_institutions.append(self.CAMPUS_INSTITUTION)

    def test_load_faculty_institution(self):
        """Test loading a faculty institution."""
        # TODO: We need to create a test MongoDB instance to test this
        # loaded = self.schema.load(self.DUMPED_FACULTY_INSTITUTION)
        # assert loaded == self.FACULTY_INSTITUTION

    def test_load_campus_institution(self):
        """Test loading a campus institution."""
        # TODO: We need to create a test MongoDB instance to test this
        # loaded = self.schema.load(self.DUMPED_CAMPUS_INSTITUTION)
        # assert loaded == self.CAMPUS_INSTITUTION

    def test_load_uni_institution(self):
        """Test loading a university institution."""
        # TODO: We need to create a test MongoDB instance to test this
        # loaded = self.schema.load(self.DUMPED_UNI_INSTITUTION)
        # assert loaded == self.UNI_INSTITUTION

    def test_dump_faculty_institution(self):
        """Test dumping a faculty institution."""
        dumped = OrderedDict(self.schema.dump(self.FACULTY_INSTITUTION))
        assert dumped == self.DUMPED_FACULTY_INSTITUTION

    def test_dump_campus_institution(self):
        """Test dumping a campus institution."""
        dumped = OrderedDict(self.schema.dump(self.CAMPUS_INSTITUTION))
        assert dumped == self.DUMPED_CAMPUS_INSTITUTION

    def test_dump_uni_institution(self):
        """Test dumping a university institution."""
        dumped = OrderedDict(self.schema.dump(self.UNI_INSTITUTION))
        assert dumped == self.DUMPED_UNI_INSTITUTION


class TestBuildingSchema:
    """Test the :class:`gator.core.schemas.institution.BuildingSchema` class.

    Class Attributes:
        BUILDINGS: A list of buildings to test with: Bahen, Robarts, and Sanford Fleming.
        DUMPED_BUILDINGS: A list of dictionaries containing the dumped building objects,
            in the same order as `BUILDINGS`.
    """

    BUILDINGS: list[models.Building] = [
        models.Building(
            code='BAH',
            name='Bahen Centre for Information Technology',
            institution=TestInstitutionSchema.CAMPUS_INSTITUTION,
            map_url='https://path.com/to/bahen/map'
        ),
        models.Building(
            code='ROB',
            name='Robarts Library',
            institution=TestInstitutionSchema.CAMPUS_INSTITUTION,
            map_url='https://path.com/to/robarts/map'
        ),
        models.Building(
            code='SF',
            name='Sanford Fleming Building',
            institution=TestInstitutionSchema.CAMPUS_INSTITUTION,
            map_url='https://path.com/to/sanford_fleming/map'
        )
    ]
    DUMPED_BUILDINGS: list[dict] = [{
        'code': building.code,
        'name': building.name,
        'institution': building.institution.code,
        'map_url': building.map_url
    } for building in BUILDINGS]

    def setup_class(self):
        """Set up the tests."""
        self.schema = schemas.BuildingSchema()

    def test_load_building(self):
        """Test loading a building."""
        # TODO: We need to create a test MongoDB instance to test this
        # for building, datum in zip(self.BUILDINGS, self.DUMPED_BUILDINGS):
        #     assert self.schema.load(datum) == building

    def test_dump_building(self):
        """Test dumping a building."""
        for building, datum in zip(self.BUILDINGS, self.DUMPED_BUILDINGS):
            assert OrderedDict(self.schema.dump(building)) == datum


class TestLocationSchema:
    """Test the :class:`gator.core.schemas.institution.LocationSchema` class.

    Class Attributes:
        LOCATIONS: A list of locations to test with: Bahen, Robarts, and Sanford Fleming.
        DUMPED_LOCATIONS: A list of dictionaries containing the dumped location objects,
            in the same order as `LOCATIONS`.
    """

    LOCATIONS: list[models.Location] = [
        models.Location(building=building, room=room)
        for building, room in zip(TestBuildingSchema.BUILDINGS, ['123', '456', '789'])
    ]
    DUMPED_LOCATIONS: list[dict] = [{
        'building': location.building.code,
        'room': location.room
    } for location in LOCATIONS]

    def setup_class(self):
        """Set up the tests."""
        self.schema = schemas.LocationSchema()

    def test_load_location(self):
        """Test loading a location."""
        # TODO: We need to create a test MongoDB instance to test this
        # for location, datum in zip(self.LOCATIONS, self.DUMPED_LOCATIONS):
        #     assert self.schema.load(datum) == location

    def test_dump_location(self):
        """Test dumping a location."""
        for location, datum in zip(self.LOCATIONS, self.DUMPED_LOCATIONS):
            assert OrderedDict(self.schema.dump(location)) == datum



"""Monkey patching for mongoengine."""
import types
from typing import Generic, Type, TypeVar

from mongoengine import Document, QuerySet

U = TypeVar('U', bound=Document)
# Register a no-op __class_getitem__ method on QuerySet to make it compatible with generic types
QuerySet.__class_getitem__ = types.MethodType(lambda self, x: self, QuerySet)


class QuerySetManager(Generic[U]):
    """A manager that returns a QuerySet of the given type."""

    def __get__(self, _: object, cls: Type[U]) -> QuerySet[U]:
        """Return a QuerySet of the given type."""
        return QuerySet(cls, cls._get_collection())



# type: ignore
"""Utilities for performing I/O operations."""
import json as jsonlib
from pathlib import Path
from typing import Any, Generator, Optional, Union

import requests


def stream_file(fp: Union[str, Path], mode: str = 'rb',
                chunk_size: int = 1024**2, **kwargs: Any) \
        -> Generator[bytes, None, None]:
    """Stream the contents of a file.

    Args:
        fp: The file path as a string or Path object.
        mode: The file mode. The default is to read the file in binary
            (rb) mode. See the `open` builtin for more information.
        chunk_size: The size of the chunks to yield. Defaults to 1 MB.
        kwargs: Additional keyword arguments passed to the `open` builtin.

    Yields:
        Chunks of size `chunk_size` representing the contents of the file.
    """
    with open(fp, mode, **kwargs) as f:
        # Stream the chunks and yield them
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            else:
                yield chunk


def http_request(url: str, method: str = 'GET', stream: bool = False,
                 chunk_size: Optional[int] = None, decode_unicode: bool = False,
                 json: bool = False, **kwargs: Any) -> Any:
    """Perform an HTTP request.

    Args:
        url: The URL to request.
        method: The HTTP method to use.
        stream: Whether to stream the response. If so, this will yield the
            chunks as they are received. Otherwise, the entire response will be
            read into memory and returned as a single chunk.
        chunk_size: The size of the chunks to yield in bytes. Only used if
            `stream` is True. Use None to yield the chunks in whatever size
            they come in.
        decode_unicode: Whether to decode the response as unicode.
        json: Whether to decode the response as JSON. If so, `stream` will be
            ignored, and the entire response will be read into memory and
            parsed as JSON.
        **kwargs: Additional keyword arguments to pass to the
            :meth:`requests.request` method. See the :mod:`requests`
            documentation for more information.

    Returns:
        The response body as a single chunk of bytes if `json` is False.
        Otherwise, the response body is parsed as JSON and returned as a
        dictionary. If `stream` is True, this will return an iterator that
        yields bytes of size `chunk_size` representing chunks of the response
        as they are received; if `chunk_size` is None, the chunks will be
        yielded in whatever size they come in.
    """
    if json:
        # Disable streaming if JSON is enabled
        stream = False

    response = requests.request(method, url, stream=stream, **kwargs)

    # Always use a chunk size of None if streaming is disabled
    chunk_size = chunk_size if stream else None
    chunks = response.iter_content(chunk_size, decode_unicode)
    if stream:
        # Return a generator that yields the chunks as they are received
        return chunks
    else:
        chunks = list(chunks)
        if json:
            # json is True, so eturn the entire response as JSON
            return jsonlib.loads(chunks[0])
        else:
            # stream is False, so return the entire response as a single chunk
            return chunks[0]



"""Marshmallow schemas for model definitions about an academic institution.

Autogenerated with marshmallow-mongoengine.
"""
from gator._vendor.marshmallow_mongoengine import ModelSchema

import gator.core.models.institution as institution


class InstitutionSchema(ModelSchema):
    """Marshmallow schema for the :class:`Institution` model."""

    class Meta:
        """Meta class for InstitutionSchema."""

        model = institution.Institution


class BuildingSchema(ModelSchema):
    """Marshmallow schema for the :class:`Building` model."""

    class Meta:
        """Meta class for BuildingSchema."""

        model = institution.Building


class LocationSchema(ModelSchema):
    """Marshmallow schema for the :class:`Location` model."""

    class Meta:
        """Meta class for LocationSchema."""

        model = institution.Location



"""Test the :mod:`gator.core.data.utils.io` module."""
import json
import math
import random
from typing import Iterator, Optional

import pytest
from pytest_httpserver import HTTPServer
from pytest_mock import MockerFixture
from werkzeug import Request, Response

from gator.core.data.utils.io import http_request, stream_file


class TestStreamFile:
    """Test the :meth:`gator.core.data.utils.io.stream_file` function.

    Class Attributes:
        EXAMPLE_FILE_DATA: The data to use for the mocked file.
    """

    EXAMPLE_FILE_DATA: bytes = b'abcdefghijklmnopqrstuvwxyz'

    @classmethod
    @pytest.fixture
    def mock_byte_file(cls, mocker: MockerFixture) -> None:
        """Fixture that mocks the `open` builtin with bytes data."""
        mocker.patch('builtins.open',
                     mocker.mock_open(read_data=cls.EXAMPLE_FILE_DATA))

    @classmethod
    @pytest.fixture
    def mock_text_file(cls, mocker: MockerFixture) -> None:
        """Fixture that mocks the `open` builtin with text data."""
        mocker.patch('builtins.open',
                     mocker.mock_open(read_data=cls.EXAMPLE_FILE_DATA.decode()))

    @pytest.mark.usefixtures('mock_byte_file')
    def test_rb_stream_1(self) -> None:
        """Test the `stream_file` function in binary read mode, with streaming.

        Use a chunk size of 2.
        """
        data = list(stream_file('test.txt', chunk_size=2))
        expected = [self.EXAMPLE_FILE_DATA[i:i + 2]
                    for i in range(0, len(self.EXAMPLE_FILE_DATA), 2)]

        assert data == expected

    @pytest.mark.usefixtures('mock_byte_file')
    def test_rb_stream_2(self) -> None:
        """Test the `stream_file` function in binary read mode, with streaming.

        Use a chunk size larger than the file size.
        """
        data = list(stream_file('test.txt',
                                chunk_size=len(self.EXAMPLE_FILE_DATA) + 1))
        assert data == [self.EXAMPLE_FILE_DATA]

    @pytest.mark.usefixtures('mock_text_file')
    def test_r_stream_1(self) -> None:
        """Test the `stream_file` function on a text file, with streaming.

        Use a chunk size of 2.
        """
        data = list(stream_file('test.txt', chunk_size=2))
        expected = [self.EXAMPLE_FILE_DATA[i:i + 2].decode('utf-8')
                    for i in range(0, len(self.EXAMPLE_FILE_DATA), 2)]

        assert data == expected

    @pytest.mark.usefixtures('mock_text_file')
    def test_r_stream_2(self) -> None:
        """Test the `stream_file` function on a text file, with streaming.

        Use a chunk size larger than the file size.
        """
        data = list(stream_file('test.txt',
                                chunk_size=len(self.EXAMPLE_FILE_DATA) + 1))
        assert data == [self.EXAMPLE_FILE_DATA.decode('utf-8')]


class TestHttpRequest:
    """Test the :meth:`gator.core.data.utils.io.http_request` function.

    Class Attributes:
        SMALL_HTML: The contents of a small page for testing.
        LARGE_HTML_SIZE: The size of the large page for testing.
        LARGE_HTML_CHAR: The character that is repeated to create the large
            page. Defaults to "a".
        LARGE_HTML_FULL_DATA: The full contents of the large page for testing.
            This is a concatenation of `LARGE_HTML_CHAR` repeated
            `LARGE_HTML_SIZE` number of times.
        EXAMPLE_JSON_DATA: The data to use for the JSON page.
        JSON_PAGE: A byte string of the `EXAMPLE_JSON_DATA` encoded as JSON.
    """

    SMALL_HTML: bytes = b"""Hello! This is a small HTML page."""
    LARGE_HTML_SIZE: int = 1024**2
    LARGE_HTML_CHAR: bytes = b'a'
    LARGE_HTML_FULL_DATA: bytes = LARGE_HTML_CHAR * LARGE_HTML_SIZE
    EXAMPLE_JSON_DATA: dict = {
        "name": "John Doe",
        "age": 42,
        "city": {"name": "New York", "state": "NY"}
    }
    JSON_PAGE: bytes = json.dumps(EXAMPLE_JSON_DATA).encode('utf-8')

    @classmethod
    def large_html_handler(cls, request: Request) -> Response:
        """Write the large html in variable sized chunks."""
        def get_chunks() -> Iterator[bytes]:
            """Generate a chunk of random bytes until max size is reached."""
            # Choose a variable sized chunk size
            total = 0
            while total < cls.LARGE_HTML_SIZE:
                chunk_size = random.randint(1, cls.LARGE_HTML_SIZE - total)
                total += chunk_size
                print(f'Chunk size: {chunk_size}')
                yield cls.LARGE_HTML_CHAR * chunk_size

        # Create a Response object from the generator
        return Response(get_chunks(),
                        mimetype='application/octet-stream',
                        headers={'Content-Length': str(cls.LARGE_HTML_SIZE)},
                        direct_passthrough=True)

    @classmethod
    @pytest.fixture
    def http_server(cls, httpserver: HTTPServer) -> HTTPServer:
        """Fixture that configures an HTTP server for testing."""
        # Configure an endpoint for the small page
        httpserver.expect_request('/small')\
            .respond_with_data(cls.SMALL_HTML)
        # Configure an endpoint for the large page
        httpserver.expect_request('/large')\
            .respond_with_handler(cls.large_html_handler)
        # Configure an endpoint for the JSON page
        httpserver.expect_request('/json')\
            .respond_with_data(cls.JSON_PAGE)
        # Configure a POST request endpoint
        httpserver.expect_request('/post', method='POST')\
            .respond_with_data(b'OK')

        return httpserver

    @pytest.mark.usefixtures('http_server')
    @pytest.mark.parametrize('chunk_size', [None, 32])
    def test_get_no_stream(self, http_server: HTTPServer,
                           chunk_size: Optional[int]) -> None:
        """Test the `http_request` function in GET mode without streaming."""
        data = http_request(http_server.url_for('/small'),
                            method='GET',
                            stream=False,
                            chunk_size=chunk_size)

        assert data == self.SMALL_HTML

    @pytest.mark.usefixtures('http_server')
    def test_get_stream_1(self, http_server: HTTPServer) -> None:
        """Test the `http_request` function in GET mode with streaming but
        no set chunk size."""
        chunks = list(http_request(http_server.url_for('/large'),
                                   method='GET',
                                   stream=True,
                                   chunk_size=None))

        # The large page is too large to fit in memory, so we expect to get
        # the chunks in a generator. Since chunk_size is None, the chunk size
        # will be variable, depending on what server is returning.
        total_size = sum(len(chunk) for chunk in chunks)
        response = b''.join(chunks)

        assert total_size == self.LARGE_HTML_SIZE
        assert response == self.LARGE_HTML_FULL_DATA

    @pytest.mark.usefixtures('http_server')
    def test_get_stream_2(self, http_server: HTTPServer) -> None:
        """Test the `http_request` function in GET mode with streaming and a
        set chunk size of 32."""
        chunks = list(http_request(http_server.url_for('/large'),
                                   method='GET',
                                   stream=True,
                                   chunk_size=32))

        # The large page is too large to fit in memory, so we expect to get
        # the chunks in batches of 32 bytes.
        num_chunks_expected = math.ceil(self.LARGE_HTML_SIZE / 32)
        assert len(chunks) == num_chunks_expected

        total_size = sum(len(chunk) for chunk in chunks)
        assert total_size == self.LARGE_HTML_SIZE

        assert chunks == [self.LARGE_HTML_FULL_DATA[i:i + 32]
                          for i in range(0, self.LARGE_HTML_SIZE, 32)]

    @pytest.mark.usefixtures('http_server')
    def test_post(self, http_server: HTTPServer) -> None:
        """Test the `http_request` function in POST mode."""
        data = http_request(http_server.url_for('/post'),
                            method='POST',
                            stream=False,
                            chunk_size=None)

        assert data == b'OK'

    @pytest.mark.usefixtures('http_server')
    def test_decode_unicode(self, http_server: HTTPServer) -> None:
        """Test the `http_request` function in GET mode without streaming and
        with decode_unicode True."""
        data = http_request(http_server.url_for('/small'),
                            method='GET',
                            stream=False,
                            chunk_size=None,
                            decode_unicode=True)

        assert data == self.SMALL_HTML.decode('utf-8')

    @pytest.mark.usefixtures('http_server')
    @pytest.mark.parametrize('stream', [False, True])
    def test_json(self, http_server: HTTPServer, stream: bool) -> None:
        """Test that the `http_request` function properly decodes JSON."""
        data = http_request(http_server.url_for('/json'),
                            method='GET',
                            stream=stream,
                            json=True)
        assert data == self.EXAMPLE_JSON_DATA



# type: ignore
"""Base classes for a datasets."""
import fnmatch
from abc import ABC, abstractclassmethod, abstractmethod, abstractproperty
from typing import Any, Iterator, Optional, Type, Union

from gator.core.models.common import Record
from gator.core.models.timetable import Session


class Dataset(ABC):
    """A dataset that returns :class:`gator.models.Record` instances.

    All subclasses should implement the `get` method to return a list of
    records. The `slug`, `name`, and `description` properties should also be
    implemented to provide metadata about the dataset.
    """

    @abstractproperty
    def slug(self) -> str:
        """Return the slug of this dataset."""
        raise NotImplementedError()

    @abstractproperty
    def name(self) -> str:
        """Return the name of this dataset."""
        raise NotImplementedError()

    @abstractproperty
    def description(self) -> str:
        """Return the description of this dataset."""
        raise NotImplementedError()

    @abstractmethod
    def get(self) -> Iterator[Record]:
        """Return an iterator that lazily yields records from this dataset."""
        raise NotImplementedError()


class SessionalDataset(Dataset):
    """A dataset that is specific to one or more sessions.

    For example, a dataset of courses offered in a set of particular sessions,
    or a dataset of clubs that are active in a single session.

    All subclasses must implement the `_get_latest_sessions` method to return
    the most up-to-date sessions for this dataset.

    Instance Attributes:
        sessions: The sessions that this dataset is specific to.
    """

    sessions: list[Session]

    def __init__(self,
                 sessions: Optional[list[Union[int, str, Session]]] = None,
                 session: Optional[Union[int, str, Session]] = None) -> None:
        """Initialize a SessionalDataset.

        Args:
            sessions: The sessions that this dataset tracks. Each element can
                either be a session code, which will be parsed into a Session
                object, or a Session object itself. If no sessions are provided,
                the most up-to-date sessions will be used.
            session: Instead of providing a list of sessions, a single session
                can be provided. This is equivalent to providing a list with a
                single session. If both `sessions` and `session` are provided,
                `sessions` will take precedence.
        """
        super().__init__()
        # Prefer the `sessions` argument. In the case it is None, use the
        # `session` argument if it is provided, otherwise default to None.
        sessions = sessions or [session] if session is not None else None
        if sessions is None:
            # Use the most up-to-date sessions
            self.sessions = self._get_latest_sessions()
        else:
            # Parse the sessions
            self.sessions = [
                Session.parse(session) if not isinstance(session, Session) else session
                for session in sessions]

    @abstractclassmethod
    def _get_latest_sessions(cls) -> list[Session]:
        """Return the most up-to-date sessions for this dataset.

        Raise a ValueError if the session could not be found.
        """
        raise NotImplementedError()


class Registry:
    """A registry tracking one or more datasets.

    Each dataset is identified by a unique slug. The registry can be used to
    retrieve datasets by their slug, or to filter datasets by a pattern.
    """

    # Private Instance Attributes:
    #     _datasets: A list of datasets to track.
    _datasets: list[Dataset]

    def __init__(self, datasets: Optional[list[Dataset]] = None) -> None:
        """Initialize a Registry.

        Args:
            datasets: A list of datasets to track.
        """
        self._datasets = datasets or []

    def register(self, dataset: Union[Type[Dataset], Dataset], **kwargs: Any) \
            -> None:
        """Register a dataset with the registry.

        Args:
            dataset: A dataset to register. This can either be a class or an
                instance of a dataset. If a class is provided, it will be
                instantiated with the provided keyword arguments.
            **kwargs: Keyword arguments to pass to the dataset initializer
                in the case that `dataset` is a class.
        """
        if isinstance(dataset, type):
            dataset = dataset(**kwargs)
        self._datasets.append(dataset)

    def filter(self, pattern: str) -> Iterator[Dataset]:
        """Filter the datasets by a pattern.

        Args:
            pattern: A Unix-style glob pattern to filter the datasets by.

        Yields:
            Datasets that match the pattern.
        """
        for dataset in self._datasets:
            if fnmatch.fnmatch(dataset.slug, pattern):
                yield dataset

    def all(self) -> Iterator[Dataset]:
        """Return an iterator that yields all datasets."""
        yield from self._datasets



"""Data hashing helper functions."""
import base64
import hashlib
from typing import Any


def make_hash_sha256(o: Any) -> str:
    """Make a hash of an object using SHA256.

    Args:
        o: The object to hash. Any object that is hashable will work.

    Returns:
        A base64 encoded string of the hash.
    """
    hasher = hashlib.sha256()
    hasher.update(repr(make_hashable(o)).encode())
    return base64.b64encode(hasher.digest()).decode()


def make_hashable(o: Any) -> Any:
    """Make an object hashable.

    If the object is a tuple or list, then it is converted to a tuple of
    hashable objects, with `make_hashable` called on each element.

    If the object is a dict, then it is converted to a sorted tuple of tuples
    where the first element is the key and the second element is the value
    of the key, with `make_hashable` called on each value.

    If the object is a set or frozenset, then it is converted to a sorted tuple
    of hashable objects, with `make_hashable` called on each element.

    Otherwise, the object is returned unchanged, with the assumption that it
    is hashable.
    """
    if isinstance(o, (tuple, list)):
        return tuple(make_hashable(e) for e in o)

    if isinstance(o, dict):
        return tuple(sorted((k, make_hashable(v)) for k, v in o.items()))

    if isinstance(o, (set, frozenset)):
        return tuple(sorted(make_hashable(e) for e in o))

    return o



"""Test the Marshmallow schemas for the timetable models."""
from collections import OrderedDict

import gator.core.models.timetable as models
import gator.core.schemas.timetable as schemas
from tests.schemas.test_institution import (TestInstitutionSchema,
                                            TestLocationSchema)

# The number of milliseconds in an hour
HOUR_MS = 60 * 60 * 1000
# The number of milliseconds in a minute
MINUTE_MS = 60 * 1000


class TestSessionSchema:
    """Test the :class:`gator.core.schemas.timetable.SessionSchema` class.

    Class Attributes:
        SESSIONS: A list of sessions to test with.
        DUMPED_SESSIONS: The dumped data for the sessions in `SESSIONS`.
    """
    SESSIONS: list[models.Session] = [
        models.Session(2020, 'fall'),
        models.Session(2021, 'winter'),
        models.Session(1966, 'summer'),
    ]

    DUMPED_SESSIONS: list[dict] = [
        {'year': session.year, 'season': session.season}  # type: ignore
        for session in SESSIONS
    ]

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.SessionSchema()

    def test_load(self) -> None:
        """Test that the schema can load a Session object."""
        for session, datum in zip(self.SESSIONS, self.DUMPED_SESSIONS):
            assert self.schema.load(datum) == session

    def test_dump(self) -> None:
        """Test that the schema can dump a Session object."""
        for session, datum in zip(self.SESSIONS, self.DUMPED_SESSIONS):
            assert OrderedDict(self.schema.dump(session)) == datum


class TestWeeklyRepetitionScheduleSchema:
    """Test the :class:`gator.core.schemas.timetable.WeeklyRepetitionScheduleSchema` class.

    Class Attributes:
        SCHEDULES: A list of schedules to test with.
        DUMPED_SCHEDULES: The dumped data for the schedules in `SCHEDULES`.
    """
    SCHEDULES = [
        models.WeeklyRepetitionSchedule(schedule=int('1', 2)),  # Every week
        models.WeeklyRepetitionSchedule(schedule=int('10', 2)),  # Every other week
        models.WeeklyRepetitionSchedule(schedule=int('101', 2)),  # Every first and third week
        models.WeeklyRepetitionSchedule(schedule=int('1010', 2)),  # Every second and fourth week
        models.WeeklyRepetitionSchedule(schedule=int('1111', 2)),  # Every week
    ]

    DUMPED_SCHEDULES = [
        {'schedule': 1, 'schedule_bits': '1', 'is_alternating': False, 'weeks': [1]},
        {'schedule': 2, 'schedule_bits': '10', 'is_alternating': True, 'weeks': [2]},
        {'schedule': 5, 'schedule_bits': '101', 'is_alternating': True, 'weeks': [1, 3]},
        {'schedule': 10, 'schedule_bits': '1010', 'is_alternating': True, 'weeks': [2, 4]},
        {'schedule': 15, 'schedule_bits': '1111', 'is_alternating': False, 'weeks': [1, 2, 3, 4]},
    ]

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.WeeklyRepetitionScheduleSchema()

    def test_load(self) -> None:
        """Test that the schema can load a WeeklyRepetitionSchedule object."""
        for schedule, datum in zip(self.SCHEDULES, self.DUMPED_SCHEDULES):
            assert self.schema.load(datum) == schedule

    def test_dump(self) -> None:
        """Test that the schema can dump a WeeklyRepetitionSchedule object."""
        for schedule, datum in zip(self.SCHEDULES, self.DUMPED_SCHEDULES):
            assert OrderedDict(self.schema.dump(schedule)) == datum


class TestSectionMeetingSchema:
    """Test the :class:`gator.core.schemas.timetable.SectionMeetingSchema` class.

    Class Attributes:
        SECTION_MEETINGS: A list of section meetings to test with.
        DUMPED_SECTION_MEETINGS: The dumped data for the section meetings in `SECTION_MEETINGS`.
    """
    # Private Class Attributes:
    #     _DUMPED_SCHEDULES: The dumped data for the schedule in each section meeting.
    _DUMPED_SCHEDULES: list[dict] = [
        TestWeeklyRepetitionScheduleSchema.DUMPED_SCHEDULES[0],
        TestWeeklyRepetitionScheduleSchema.DUMPED_SCHEDULES[0],
        TestWeeklyRepetitionScheduleSchema.DUMPED_SCHEDULES[1]
    ]

    SECTION_MEETINGS = [
        # Monday 12:30-13:30 in ABC 123 for 2020 fall
        models.SectionMeeting(
            day=0,
            start_time=12 * HOUR_MS + 30 * MINUTE_MS,
            end_time=13 * HOUR_MS + 30 * MINUTE_MS,
            session=TestSessionSchema.SESSIONS[0],  # Fall 2020
            location=TestLocationSchema.LOCATIONS[0]
        ),
        # Tuesday 14:00-15:00 in DEF 456 for 2021 winter
        models.SectionMeeting(
            day=1,
            start_time=14 * HOUR_MS,
            end_time=15 * HOUR_MS,
            session=TestSessionSchema.SESSIONS[1],  # Winter 2021
            location=TestLocationSchema.LOCATIONS[1]
        ),
        # Friday 09:00-12:00 in GHI 789 for 1966 summer, every other week
        models.SectionMeeting(
            day=4,
            start_time=9 * HOUR_MS,
            end_time=12 * HOUR_MS,
            session=TestSessionSchema.SESSIONS[2],  # Summer 1966
            location=TestLocationSchema.LOCATIONS[2],
            repetition_schedule=TestWeeklyRepetitionScheduleSchema.SCHEDULES[1]  # Every other week
        )
    ]

    DUMPED_SECTION_MEETINGS = [
        {
            'day': section.day,  # type: ignore
            'start_time': section.start_time,  # type: ignore
            'end_time': section.end_time,  # type: ignore
            'session': dumped_session,
            'location': dumped_location,
            'repetition_schedule': dumped_schedule
        }
        for section, dumped_session, dumped_location, dumped_schedule in zip(
            SECTION_MEETINGS,
            TestSessionSchema.DUMPED_SESSIONS,
            TestLocationSchema.DUMPED_LOCATIONS,
            _DUMPED_SCHEDULES
        )
    ]

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.SectionMeetingSchema()

    def test_load(self) -> None:
        """Test that the schema can load a SectionMeeting object."""
        # TODO: We need to create a test MongoDB instance to test this
        # for section, datum in zip(self.SECTION_MEETINGS, self.DUMPED_SECTION_MEETINGS):
        #     assert self.schema.load(datum) == section

    def test_dump(self) -> None:
        """Test that the schema can dump a SectionMeeting object."""
        for section, datum in zip(self.SECTION_MEETINGS, self.DUMPED_SECTION_MEETINGS):
            assert OrderedDict(self.schema.dump(section)) == datum


class TestInstructorSchema:
    """Test the :class:`gator.core.schemas.timetable.InstructorSchema` class.

    Class Attributes:
        INSTRUCTORS: A list of instructors to test with.
        DUMPED_INSTRUCTORS: The dumped data for the instructors in `INSTRUCTORS`.
    """
    INSTRUCTORS: list[models.Instructor] = [
        models.Instructor(first_name='Jane', last_name='Doe'),
        models.Instructor(first_name='John', last_name='Doe'),
        models.Instructor(first_name='John', last_name='Smith'),
    ]
    DUMPED_INSTRUCTORS = [{
        'first_name': instructor.first_name,
        'last_name': instructor.last_name
    } for instructor in INSTRUCTORS]

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.InstructorSchema()

    def test_load(self) -> None:
        """Test that the schema can load an Instructor object."""
        for instructor, datum in zip(self.INSTRUCTORS, self.DUMPED_INSTRUCTORS):
            assert self.schema.load(datum) == instructor

    def test_dump(self) -> None:
        """Test that the schema can dump an Instructor object."""
        for instructor, datum in zip(self.INSTRUCTORS, self.DUMPED_INSTRUCTORS):
            assert OrderedDict(self.schema.dump(instructor)) == datum


class TestEnrolmentInfoSchema:
    """Test the :class:`gator.core.schemas.timetable.EnrolmentInfoSchema` class.

    Class Attributes:
        ENROLMENT_INFO: An enrolment info to test with.
        DUMPED_ENROLMENT_INFO: The dumped data for `ENROLMENT_INFO`.
    """
    ENROLMENT_INFO: models.EnrolmentInfo = models.EnrolmentInfo(
        current_enrolment=10,
        max_enrolment=20,
        has_waitlist=True,
        current_waitlist_size=5
    )
    DUMPED_ENROLMENT_INFO = {
        'current_enrolment': ENROLMENT_INFO.current_enrolment,
        'max_enrolment': ENROLMENT_INFO.max_enrolment,
        'has_waitlist': ENROLMENT_INFO.has_waitlist,
        'current_waitlist_size': ENROLMENT_INFO.current_waitlist_size
    }

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.EnrolmentInfoSchema()

    def test_load(self) -> None:
        """Test that the schema can load an EnrolmentInfo object."""
        assert self.schema.load(self.DUMPED_ENROLMENT_INFO) == self.ENROLMENT_INFO

    def test_dump(self) -> None:
        """Test that the schema can dump an EnrolmentInfo object."""
        assert OrderedDict(self.schema.dump(self.ENROLMENT_INFO)) == self.DUMPED_ENROLMENT_INFO


class TestSectionSchema:
    """Test the :class:`gator.core.schemas.timetable.SectionSchema` class.

    Class Attributes:
        SECTION: A section to test with.
        DUMPED_SECTION: The dumped data for `SECTION`.
    """
    SECTION: models.Section = models.Section(
        teaching_method=models.TeachingMethod.LECTURE,
        section_number="0101",
        meetings=TestSectionMeetingSchema.SECTION_MEETINGS,
        instructors=TestInstructorSchema.INSTRUCTORS,
        delivery_modes=[
            models.SectionDeliveryMode.IN_PERSON,  # Fall 2020
            models.SectionDeliveryMode.IN_PERSON   # Winter 2021
        ],
        enrolment_info=TestEnrolmentInfoSchema.ENROLMENT_INFO
    )
    DUMPED_SECTION = {
        'teaching_method': SECTION.teaching_method.name,
        'section_number': SECTION.section_number,
        'meetings': TestSectionMeetingSchema.DUMPED_SECTION_MEETINGS,
        'instructors': TestInstructorSchema.DUMPED_INSTRUCTORS,
        'delivery_modes': [mode.name for mode in SECTION.delivery_modes],
        'cancelled': SECTION.cancelled,
        'enrolment_info': TestEnrolmentInfoSchema.DUMPED_ENROLMENT_INFO,
    }

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.SectionSchema()

    def test_load(self) -> None:
        """Test that the schema can load a Section object."""
        # TODO: We need to create a test MongoDB instance to test this
        # assert self.schema.load(self.DUMPED_SECTION) == self.SECTION

    def test_dump(self) -> None:
        """Test that the schema can dump a Section object."""
        assert OrderedDict(self.schema.dump(self.SECTION)) == self.DUMPED_SECTION


class TestCategoricalRequirementSchema:
    """Test the :class:`gator.core.schemas.timetable.CategoricalRequirementSchema` class.

    Class Attributes:
        CATEGORICAL_REQUIREMENT: A categorical requirement to test with.
        DUMPED_CATEGORICAL_REQUIREMENT: The dumped data for `CATEGORICAL_REQUIREMENT`.
    """
    CATEGORICAL_REQUIREMENT: models.CategoricalRequirement = models.CategoricalRequirement(
        code='BR=1',
        type='breadth',
        name='Creative Cultural',
        description='BR=1 Creative and Cultural Representation',
        institutions=[TestInstitutionSchema.FACULTY_INSTITUTION]
    )
    DUMPED_CATEGORICAL_REQUIREMENT = {
        'code': CATEGORICAL_REQUIREMENT.code,
        'type': CATEGORICAL_REQUIREMENT.type,
        'name': CATEGORICAL_REQUIREMENT.name,
        'description': CATEGORICAL_REQUIREMENT.description,
        'institutions': [TestInstitutionSchema.FACULTY_INSTITUTION.code]
    }

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.CategoricalRequirementSchema()

    def test_load(self) -> None:
        """Test that the schema can load a CategoricalRequirement object."""
        # TODO: We need to create a test MongoDB instance to test this
        # assert self.schema.load(self.DUMPED_CATEGORICAL_REQUIREMENT) == self.CATEGORICAL_REQUIREMENT

    def test_dump(self) -> None:
        """Test that the schema can dump a CategoricalRequirement object."""
        assert OrderedDict(self.schema.dump(self.CATEGORICAL_REQUIREMENT)) == self.DUMPED_CATEGORICAL_REQUIREMENT


class TestCourseSchema:
    """Test the :class:`gator.core.schemas.timetable.CourseSchema` class.

    Class Attributes:
        COURSE: A course to test with.
        DUMPED_COURSE: The dumped data for `COURSE`.
    """
    COURSE: models.Course = models.Course(
        id='6e1c6b5c-1b1d-4b1e-9c1c-1b1d4b1e9c1c',
        code='XYZ300Y1',
        name='Introduction to XYZ',
        sections=[TestSectionSchema.SECTION],
        sessions=[TestSessionSchema.SESSIONS[0], TestSessionSchema.SESSIONS[1]],  # Fall 2020, Winter 2021
        term=models.Term.FULL_YEAR,
        credits=1.0,
        institution=models.Institution(
            code='XYZ',
            name='XYZ Department',
            type='department',
            parent=TestInstitutionSchema.FACULTY_INSTITUTION
        ),
        instruction_level=models.InstructionLevel.UNDERGRADUATE,
        description='This course introduces students to XYZ.',
        categorical_requirements=[TestCategoricalRequirementSchema.CATEGORICAL_REQUIREMENT],
        prerequisites='XYZ100Y1',
        corequisites='XYZ200Y1',
        exclusions='XYZ400Y1',
        cancelled=False,
        tags=['xyz', 'introductory', 'third-year'],
        notes=['This course is only offered for students in the XYZ program.']
    )
    DUMPED_COURSE = {
        'id': COURSE.id,
        'code': COURSE.code,
        'name': COURSE.name,
        'sections': [TestSectionSchema.DUMPED_SECTION],
        'sessions': [dumped_session for dumped_session in TestSessionSchema.DUMPED_SESSIONS[:2]],
        'term': COURSE.term.name,
        'credits': COURSE.credits,
        'institution': COURSE.institution.code,
        'instruction_level': COURSE.instruction_level.name,  # type: ignore
        'description': COURSE.description,
        'categorical_requirements': [TestCategoricalRequirementSchema.DUMPED_CATEGORICAL_REQUIREMENT],
        'prerequisites': COURSE.prerequisites,
        'corequisites': COURSE.corequisites,
        'exclusions': COURSE.exclusions,
        'cancelled': COURSE.cancelled,
        'tags': COURSE.tags,
        'notes': COURSE.notes
    }

    def setup_class(self) -> None:
        """Initialize the test class."""
        self.schema = schemas.CourseSchema()

    def test_load(self) -> None:
        """Test that the schema can load a Course object."""
        # TODO: We need to create a test MongoDB instance to test this
        # assert self.schema.load(self.DUMPED_COURSE) == self.COURSE

    def test_dump(self) -> None:
        """Test that the schema can dump a Course object."""
        assert OrderedDict(self.schema.dump(self.COURSE)) == self.DUMPED_COURSE



"""Model definitions for timetable data."""
import math
import re
from functools import lru_cache
from typing import Any, Optional, Union

from mongoengine import Document, EmbeddedDocument, fields

from gator.core.models.common import SerializableEnum
from gator.core.models.institution import Institution, Location


class Session(EmbeddedDocument):
    """A formal division of the academic year.

    The regular academic year is divided into three sessions: the fall session,
    which runs from September to December; the winter (spring) session, which
    runs from January to April; and the summer session, which runs from May to
    August.

    The canonical representation of a session is a five-length string where the
    first four characters denote the session year, and the last character denotes
    the session season, 9 for fall, 1 for winter, and 5 for summer. For example,
    the code `20209` denotes the fall session of 2020, and the code `19665`
    denotes the summer session of 1966.

    Instance Attributes:
        year: The year of the session. Must be between 0 and 9999.
        season: The season of the session.
    """

    # Private Class Attributes:
    #     _SEASON_MAP: A mapping from season names to their corresponding
    #         session codes.
    #     _INVERSE_SEASON_MAP: A mapping from session codes to their
    #         corresponding season names.
    _SEASON_MAP: dict[str, int] = {'fall': 9, 'winter': 1, 'summer': 5}
    _INVERSE_SEASON_MAP: dict[int, str] = {v: k for k, v in _SEASON_MAP.items()}

    year: int = fields.IntField(required=True, min_value=0, max_value=9999)  # type: ignore
    season: str = fields.StringField(required=True, choices=list(_SEASON_MAP.keys()))  # type: ignore

    def __init__(self, year: int, season: str, *args: Any, **kwargs: Any) -> None:
        """Initialize a session with the given year and season.

        Args:
            year: The year of the session. Must be between 0 and 9999.
            season: The season of the session.

        Note:
            The year and season are passed as positional arguments for
            convenience, but they are actually stored as keyword arguments
            in the `kwargs` dictionary. If `year` and `season` are passed
            as keyword arguments, they will be ignored, meaning that the
            positional arguments always take precedence.
        """
        kwargs['year'] = year
        kwargs['season'] = season
        super().__init__(*args, **kwargs)

    @property
    def code(self) -> str:
        """The canonical representation of this session.

        Returns:
            A five-length string where the first four characters denote the
            session year, and the last character denotes the session season,
            9 for fall, 1 for winter, and 5 for summer.

        Examples:
            >>> Session(2020, 'fall').code
            '20209'
            >>> Session(2019, 'winter').code
            '20191'
            >>> Session(1966, 'summer').code
            '19665'
        """
        return f'{str(self.year).zfill(4)}{self._SEASON_MAP[self.season]}'

    @property
    def human_str(self) -> str:
        """A human-readable representation of this session.

        Examples:
            >>> Session(2020, 'fall').human_str
            'Fall 2020'
            >>> Session(2019, 'winter').human_str
            'Winter 2019'
            >>> Session(1966, 'summer').human_str
            'Summer 1966'
        """
        return f'{self.season.capitalize()} {self.year}'

    def __str__(self) -> str:
        """Return a string representation of this session.

        The output string is the session code. See `code` for more
        information.
        """
        return self.code

    @classmethod
    def parse(cls, session_code: Union[str, int]) -> 'Session':
        """Return an instance Session representing the given session code.

        Raise a ValueError if the session code is not formatted properly.

        Args:
            session_code: The session code to parse as a string or integer.
                If the session code is an integer, it will be converted to a
                string and zero-padded to the left if necessary. E.g. if
                509 is passed, it will be converted to '00509' which corresponds
                to the fall session of the year 50.

        Examples:
            >>> Session.parse(509) == Session(year=50, season='fall')
            True
            >>> Session.parse(20201) == Session(year=2020, season='winter')
            True
            >>> Session.parse('20205') == Session(year=2020, season='summer')
            True
        """
        # Ensure that the session code is a string and zero-padded to the left
        session_code = str(session_code).zfill(5)

        if len(session_code) != 5:
            raise ValueError(
                f'invalid session code ("{session_code}"): expected string of length 5'
                f', not {len(session_code)}')
        elif not session_code.isnumeric():
            raise ValueError(
                f'invalid session code ("{session_code}"): expected numeric string')
        elif int(session_code[-1]) not in cls._INVERSE_SEASON_MAP:
            raise ValueError(
                f'invalid session code ("{session_code}"): expected code to end in '
                f'one of {list(cls._INVERSE_SEASON_MAP.keys())}, not {session_code[-1]}')
        else:
            return Session(
                int(session_code[:4]),
                cls._INVERSE_SEASON_MAP[int(session_code[-1])])


class WeeklyRepetitionSchedule(EmbeddedDocument):
    """A schedule for an event that repeats weekly.

    Represented as an integer whose binary representation is
    interpreted as a set of :math:`k` bits, where :math:`k` is the number of
    weeks in a single repeating period. The :math:`i`-th bit is 1 if the
    meeting occurs in the :math:`i`-th week of the repeating period, and 0
    otherwise. For example, if a meeting occurs every week, then :math:`k = 1`
    and the bit is `1`. If a meeting occurs every other week, then
    :math:`k = 2` and the bits are `10`.

    Instance Attributes:
        schedule: The integer representing the repetition schedule.
    """

    schedule: int = fields.IntField(min_value=1, default=1)  # type: ignore

    @property
    def is_alternating(self) -> bool:
        """Whether this meeting does not regularly occur every week.

        A schedule is alternating (i.e. does not occur every week) if there is
        at least one week :math:`i` such that the :math:`i`-th bit is 0.

        Examples:
            >>> WeeklyRepetitionSchedule(schedule=0b1010).is_alternating
            True
            >>> WeeklyRepetitionSchedule(schedule=0b1111).is_alternating
            False
        """
        k = math.ceil(math.log2(self.schedule))
        return self.schedule > 1 and self.schedule != (1 << k) - 1

    @property
    def schedule_bits(self) -> str:
        """The binary representation of the schedule.

        Examples:
            >>> WeeklyRepetitionSchedule(schedule=0b1).schedule_bits
            '1'
            >>> WeeklyRepetitionSchedule(schedule=0b1010).schedule_bits
            '1010'
        """
        return bin(self.schedule)[2:]  # Remove the '0b' prefix

    @property
    def weeks(self) -> list[int]:
        """A list of the weeks in which this meeting occurs.

        Each week is represented as an integer between 1 and the number of
        weeks in a repeating period, inclusive. The weeks are sorted in
        ascending order.

        Examples:
            >>> WeeklyRepetitionSchedule(schedule=0b1).weeks
            [1]
            >>> WeeklyRepetitionSchedule(schedule=0b1010).weeks
            [2, 4]
        """
        return [i + 1 for i, bit in enumerate(reversed(self.schedule_bits)) if bit == '1']


class SectionMeeting(EmbeddedDocument):
    """A single meeting of a section.

    Instance Attributes:
        day: The day of the week that this meeting occurs on, represented as an
            integer, where Monday is 0 and Sunday is 6.
        start_time: The time that this meeting starts, represented as the number
            of milliseconds since midnight(00:00:00).
        end_time: The time that this meeting ends, represented as the number of
            milliseconds since midnight (00:00:00).
        session: The session that this meeting occurs in.
        location: The location of this meeting, or None if this meeting has no
            assigned location.
        repetition_schedule: The repetition schedule of this meeting. Defaults
            to a schedule that occurs every week.
    """

    day: int = fields.IntField(required=True, min_value=0, max_value=6)  # type: ignore
    start_time: int = fields.IntField(required=True, min_value=0, max_value=86400000)  # type: ignore
    end_time: int = fields.IntField(required=True, min_value=0, max_value=86400000)  # type: ignore
    session: Session = fields.EmbeddedDocumentField(Session, required=True)  # type: ignore
    location: Location = fields.EmbeddedDocumentField(Location, null=True, default=None)  # type: ignore
    repetition_schedule: WeeklyRepetitionSchedule = fields.EmbeddedDocumentField(
        WeeklyRepetitionSchedule, default=WeeklyRepetitionSchedule)  # type: ignore


class TeachingMethod(SerializableEnum):
    """A class representing the teaching method for a section."""

    LECTURE = 'LEC'
    TUTORIAL = 'TUT'
    PRACTICAL = 'PRA'


class SectionDeliveryMode(SerializableEnum):
    """A class representing mode of delivery for a section."""

    CLASS = 'CLASS'
    ONLINE_SYNC = 'ONLSYNC'
    ONLINE_ASYNC = 'ONLASYNC'
    IN_PERSON = 'INPER'
    SYNC = 'SYNC'
    ASYNC = 'ASYNC'
    # idk why they keep changing delivery modes
    ASYIF = 'ASYIF'
    SYNIF = 'SYNIF'


class Instructor(EmbeddedDocument):
    """A course instructor.

    Instance Attributes:
        first_name: The first name of this instructor.
        last_name: The last name of this instructor.
    """

    first_name: str = fields.StringField(required=True)  # type: ignore
    last_name: str = fields.StringField(required=True)  # type: ignore


class EnrolmentInfo(EmbeddedDocument):
    """Enrolment information for a section.

    Instance Attributes:
        current_enrolment: The number of students currently enrolled in this
            section.
        max_enrolment: The maximum number of students that can be enrolled in
            this section.
        has_waitlist: Whether this section has a waitlist.
        current_waitlist_size: The number of students currently on the waitlist
            for this section.
        enrolment_indicator: A string representing the enrollment indicator
            for this section, or None if there is no enrollment indicator.
    """

    current_enrolment: Optional[int] = fields.IntField(null=True, min_value=0, default=None)  # type: ignore
    max_enrolment: Optional[int] = fields.IntField(null=True, min_value=0, default=None)  # type: ignore
    has_waitlist: bool = fields.BooleanField(default=False)  # type: ignore
    current_waitlist_size: Optional[int] = fields.IntField(null=True, min_value=0, default=None)  # type: ignore
    enrolment_indicator: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    # TODO: Add enrolment controls


class Section(EmbeddedDocument):
    """A section of a course.

    Instance Attributes:
        teaching_method: The teaching method for this section.
        section_number: The number of this section, represented as a string.
        meetings: A list of meetings available for this section.
        instructors: A list of instructors teaching this section.
        delivery_modes: A list of delivery modes for this section. The i-th
            element of this list corresponds to the delivery mode in the
            i-th term that the course runs in.
        subtitle: The subtitle of this section, or None if there is no subtitle.
            Among other things, this is used to distinguish between the content
            matter when a course offers curriculum on multiple topics. For
            example, an introductory algorithms course might have a section
            that uses Python and another section that uses Java.
        cancelled: Whether this section is cancelled.
        enrolment_info: The enrolment information for this section.
        notes: A list of HTML strings.
    """

    teaching_method: TeachingMethod = fields.EnumField(TeachingMethod, required=True)  # type: ignore
    section_number: str = fields.StringField(required=True)  # type: ignore
    meetings: list[SectionMeeting] = fields.EmbeddedDocumentListField('SectionMeeting', required=True)  # type: ignore
    instructors: list[Instructor] = fields.EmbeddedDocumentListField('Instructor', required=True)  # type: ignore
    delivery_modes: list[SectionDeliveryMode] = fields.ListField(
        fields.EnumField(SectionDeliveryMode), required=True)  # type: ignore
    subtitle: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    cancelled: bool = fields.BooleanField(default=False)  # type: ignore
    enrolment_info: EnrolmentInfo = fields.EmbeddedDocumentField(EnrolmentInfo, default=EnrolmentInfo)  # type: ignore
    notes: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore
    # TODO: Section dependencies. This should be represented as a graph, where
    # each node is a section and each edge is a dependency (which may or may not
    # be bidirectional).
    linked_sections: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore

    @property
    def code(self) -> str:
        """Return a string representing the code of this section.

        This is a combination of the teaching method and the section
        number, separated by a hyphen.

        If the teaching method is None, then the section number is returned
        instead.
        """
        if self.teaching_method is None:
            return self.section_number
        else:
            return f'{self.teaching_method.value}-{self.section_number}'


class Term(SerializableEnum):
    """The term in which a course is offered.

    Class Attributes:
        FIRST_SEMESTER: The first semester of the academic year.
        SECOND_SEMESTER: The second semester of the academic year.
        FULL_YEAR: The full academic year. Spans both semesters.
    """

    FIRST_SEMESTER = 'F'
    SECOND_SEMESTER = 'S'
    FULL_YEAR = 'Y'


class InstructionLevel(SerializableEnum):
    """The level of instruction for a course."""

    UNDERGRADUATE = 'undergraduate'


class CategoricalRequirement(EmbeddedDocument):
    """A breadth or distribution requirement.

    Instance Attributes:
        code: A unique code for this categorical requirement.
        type: The type of this categorical requirement. Either 'breadth' or
            'distribution'.
        name: The name of this categorical requirement.
        description: A description of this categorical requirement.
        institutions: A list of institutions where this categorical requirement
            applies. For example, a breadth requirement might only apply to
            students in the Faculty of Arts and Science, while a distribution
            requirement might apply to all students. There should be at least
            one institution in this list.
    """

    code: str = fields.StringField(unique=True, primary_key=True)  # type: ignore
    type: str = fields.StringField(required=True, choices=['breadth', 'distribution'])  # type: ignore
    name: str = fields.StringField()  # type: ignore
    description: str = fields.StringField()  # type: ignore
    institutions: list[Institution] = fields.ListField(
        fields.ReferenceField(Institution), required=True)  # type: ignore


class Course(Document):
    """An instance of a course offered at UofT for a specific term.

    Instance Attributes:
        id: A unique identifier for this course.
        code: The code of this course.
        name: The name of this course.
        sections: A list of sections available for this course.
        sessions: A list of sessions in which this course is offered.
        term: The term this course is offered in.
        credits: The number of credits this course is worth.
        institution: The institution that offers this course. This is a sub-tree
            of the institution hierarchy.
        title: The title of this course. This is usually the same as the name.
        instruction_level: The level of instruction for this course.
            For example, undergraduate or graduate.
        categorical_requirements: Breadth and distribution requirements that this
            course fulfills.
        description: The description of this course.
        prerequisites: Prerequisties for this course.
        corequisites: Corequisites for this course.
        exclusions: Exclusions for this course.
        recommended_preparation: Recommended preparations to complete before this course.
        cancelled: Whether this course is cancelled.
        tags: Tags for this course.
        notes: A list of HTML strings.
    """

    id: str = fields.StringField(primary_key=True)  # type: ignore
    code: str = fields.StringField()  # type: ignore
    name: str = fields.StringField()  # type: ignore
    sections: list[Section] = fields.EmbeddedDocumentListField('Section')  # type: ignore
    sessions: list[Session] = fields.EmbeddedDocumentListField('Session')  # type: ignore
    term: Term = fields.EnumField(Term)  # type: ignore
    credits: float = fields.FloatField()  # type: ignore
    institution: Institution = fields.ReferenceField(Institution)  # type: ignore
    # Metadata fields
    title: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    instruction_level: Optional[InstructionLevel] = fields.EnumField(
        InstructionLevel, null=True, default=None)  # type: ignore
    description: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    categorical_requirements: list[CategoricalRequirement] = fields.EmbeddedDocumentListField(
        CategoricalRequirement, default=list)  # type: ignore
    prerequisites: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    corequisites: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    exclusions: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    recommended_preparation: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    cancelled: bool = fields.BooleanField(default=False)  # type: ignore
    tags: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore
    notes: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore

    @property
    @lru_cache
    def section_codes(self) -> set[str]:
        """Return a set of section codes for this course."""
        return {section.code for section in self.sections}

    @property
    @lru_cache
    def level(self) -> int:
        """Return the level of this course.

        If the course is not a level course, return 0.
        """
        m = re.search(r'(?:[^\d]*)(\d+)', self.code)
        if not m:
            return 0
        else:
            return int(math.floor(int(m.group(1)) / 100.0)) * 100



"""Shared models used by the API."""
from datetime import datetime
from enum import Enum
from typing import Any

from mongoengine import Document, fields

from gator.core.models.mongoengine_typing import QuerySetManager


class SerializableEnum(Enum):
    """An enum that can be serialized to a JSON object."""

    def __str__(self):
        """Return the name of the enum."""
        return self.name


class Record(Document):
    """A class representing a record.

    Instance Attributes:
        id: The unique ID of the record. This should be consistent across all
            changes made to the underlying data.
        doc: The underlying data of the record.
        created_at: The time when the record was created.
        updated_at: The time when the record was last updated.
        hash: The hash of the underlying data.
        name: A short string that gives a human-readable name for the record.
            This is used for display purposes only. If not provided, the ID
            will be used.
    """

    id: str = fields.StringField(primary_key=True)  # type: ignore
    doc: Any = fields.GenericReferenceField(required=True)  # type: ignore
    created_at: datetime = fields.DateTimeField(required=True, default=datetime.now)  # type: ignore
    updated_at: datetime = fields.DateTimeField(required=True, default=datetime.now)  # type: ignore
    hash: str = fields.StringField(required=True, unique=True)  # type: ignore
    name: str = fields.StringField(required=False)  # type: ignore
    objects: QuerySetManager['Record'] = QuerySetManager['Record']()

    def sync(self, force: bool = False) -> str:
        """Sync the record with the database.

        Args:
            force: If True, force the sync even if the data is already up-to-date.

        Returns:
            One of 'created', 'updated', or 'skipped'.
        """
        # TODO: Better syncing logic that can handle batch updates

        # Check if the record is already in the database
        record = Record.objects(id=self.id).first()
        if record is None:
            # Create a new record
            self.save(cascade=True)
            return 'created'
        elif force or record.hash != self.hash:
            # Update the record
            record.doc = self.doc
            record.updated_at = datetime.now()
            record.hash = self.hash
            record.save(cascade=True)
            return 'updated'
        else:
            # Skip the record
            return 'skipped'



"""Model definitions for data about an academic institution."""
from queue import Queue
from typing import Optional

import mongoengine
from mongoengine import Document, EmbeddedDocument, fields


class Institution(Document):
    """An academic institution.

    This is represented as a recursive tree structure, where each institution
    can have multiple sub-institutions. For example, the University of Toronto
    is the parent institution of the St. George campus of the University of
    Toronto, which is the parent institution of many faculties and departments.

    Instance Attributes:
        code: A unique string representing this institution.
        name: The full name of this institution.
        type: The type of this institution. For example, `university` or
            `campus`.
        parent: The parent institution of this institution. For example, the
            University of Toronto is the parent institution of the St. George
            campus of the University of Toronto. If None, this institution is
            a top-level institution.
        sub_institutions: A list of institutions that are sub-institutions of
            this institution. For example, the St. George campus of the
            University of Toronto is a sub-institution of the University of
            Toronto.
    """

    code: str = fields.StringField(unique=True, primary_key=True)  # type: ignore
    name: str = fields.StringField(required=True)  # type: ignore
    type: str = fields.StringField(required=True)  # type: ignore
    # Institution hierarchy
    parent: Optional['Institution'] = fields.ReferenceField('self', default=None, reverse_delete_rule=mongoengine.NULLIFY)  # type: ignore
    sub_institutions: list['Institution'] = fields.ListField(
        fields.ReferenceField('self'), default=list, reverse_delete_rule=mongoengine.NULLIFY)  # type: ignore

    def find_children(self, institution_type: str) -> list['Institution']:
        """Find all sub-institutions of the given type.

        This function recursively searches through all sub-institutions of this
        institution, performing a queue-based breadth-first search, and returns
        a list of all institutions of the given type.

        Args:
            institution_type: The type of institution to search for. For
                example, `university` or `campus`.

        Returns:
            A list of institutions of the given type.
        """
        q = Queue()
        q.put(self)

        visited = set()
        children = []

        while not q.empty():
            curr = q.get()
            if curr.code in visited:
                continue
            visited.add(curr.code)

            if curr.type == institution_type:
                children.append(curr)

            for sub_inst in curr.sub_institutions:
                q.put(sub_inst)

        return children

    def find_parent(self, institution_type: str) -> Optional['Institution']:
        """Find the first parent institution of the given type.

        This function recursively searches through all parent institutions of
        this institution, performing a queue-based breadth-first search, and
        returns the first institution of the given type.

        Args:
            institution_type: The type of institution to search for. For
                example, `university` or `campus`.

        Returns:
            The first (closest) parent institution of the given type, or None
            if no such institution exists.
        """
        q = Queue()
        q.put(self)

        visited = set()

        while not q.empty():
            curr = q.get()
            if curr.code in visited:
                continue
            visited.add(curr.code)

            if curr.type == institution_type:
                return curr

            if curr.parent is not None:
                q.put(curr.parent)

        return None


class Building(Document):
    """A building on a campus of the University of Toronto.

    Instance Attributes:
        code: A unique string code that identifies this building.
        institution: The institution that this building belongs to.
        name: The name of this building. Can be None if unknown.
        map_url: The URL of this building's map. Can be None if unknown.
    """

    code: str = fields.StringField(unique=True, primary_key=True)  # type: ignore
    institution: Institution = fields.ReferenceField(Institution, required=True)  # type: ignore
    name: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    map_url: Optional[str] = fields.URLField(null=True, default=None)  # type: ignore


class Location(EmbeddedDocument):
    """A location on campus represented as a building-room pair.

    Instance Attributes:
        building: The building that the section meets in.
        room: The room that the section meets in, not including the building
            code.
    """

    building: Building = fields.ReferenceField(Building, required=True)  # type: ignore
    room: str = fields.StringField(required=True)  # type: ignore
